//
// $Id: Empire.h,v 1.13 1997/10/31 04:51:47 nygard Exp $
//

//
//  This file is a part of Empire, a game of exploration and conquest.
//  Copyright (C) 1996  Steve Nygard
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  You may contact the author by:
//     e-mail:  nygard@telusplanet.net
//

#import <Foundation/Foundation.h>

#import "SNUtility.h"
#import "NSObject-SNExtensions.h"

//----------------------------------------------------------------------
// There are three valid types of terrain.  The "unknown" type isn't
// really used -- a separate explored flag is used to hide unexplored
// terrain.
//----------------------------------------------------------------------

typedef enum _Terrain
{
    t_unknown = 0,
    t_water,
    t_land,
    t_city
} Terrain;

typedef enum _Player
{
    p_neutral = 0,
    p_player1,
    p_player2,
    p_player3
} Player;

//----------------------------------------------------------------------
// An Icon is the on screen representation of a unit.  Some unit types
// have more than one Icon to represent their different states:
//
//               Normal                 Other
//   army:       i_army                 i_sentry
//   transport:  i_unloaded_transport   i_loaded_transport
//   carrier:    i_unloaded_carrier     i_loaded_carrier
//----------------------------------------------------------------------

typedef enum _Icon
{
    i_none = 0,
    i_army,
    i_sentry,
    i_fighter,
    i_unloaded_transport,
    i_loaded_transport,
    i_submarine,
    i_destroyer,
    i_cruiser,
    i_unloaded_carrier,
    i_loaded_carrier,
    i_battleship,
    i_hovercraft
} Icon;

//----------------------------------------------------------------------
// These represent the different types of units that are available.
// The hovercraft type isn't part of the original game.
//----------------------------------------------------------------------

typedef enum _UnitType
{
    u_unknown = 0,
    u_army,
    u_fighter,
    u_transport,
    u_submarine,
    u_destroyer,
    u_cruiser,
    u_carrier,
    u_battleship,
    u_hovercraft
} UnitType;

#define UNIT_TYPE_COUNT 10

//----------------------------------------------------------------------
// A MapToken encodes four things:
//
//   - the background terrain
//   - the player number of the city or unit occupying this location
//   - the icon of the unit occupying this location, if any
//   - an explored flag.  Unexplored territory will show up black.
//----------------------------------------------------------------------

typedef int MapToken;

//----------------------------------------------------------------------
// The bitfields of a MapToken:
// e.p p.i i i i.t t
//
// e - explored  set if UNexplored
// p - player
// i - icon
// t - terrain
//----------------------------------------------------------------------

#define MT_EXPLORED_MASK 0x100
#define MT_PLAYER_MASK   0x0c0
#define MT_ICON_MASK     0x03c
#define MT_TERRAIN_MASK  0x003

typedef enum _Direction
{
    d_northwest,
    d_north,
    d_northeast,
    d_west,
    d_none,
    d_east,
    d_southwest,
    d_south,
    d_southeast
} Direction;

typedef enum _MoveResult
{
    mr_zero_range,     // Has already moved - no range left for this turn.
    mr_moved,          // Moved into empty space.
    mr_blocked,        // Unit can't move onto terrain, or terrain occupied by own unit.
    mr_victory,        // Attacked an enemy unit/city, won and moved into space.
    mr_captured_city,  // Captured enemy city (with an army, which is disbanded for city garrison).
    mr_destroyed,      // Attacked/bombarded an enemy unit/city, and lost.
} MoveResult;

typedef struct _EMMapLocation
{
    int row;
    int column;
} EMMapLocation;

typedef struct _EMMapSize
{
    int width;
    int height;
} EMMapSize;

//----------------------------------------------------------------------
// This encodes an array of nine map tokens, which come from a 3 by 3
// area of the map.  This structure is used because arrays cannot be
// passed over Distributed Objects.
//----------------------------------------------------------------------

struct NineTokens
{
    MapToken tokens[9];
};

//----------------------------------------------------------------------
// A CombatProfile encodes everything we need to know about an attacking
// or defending unit, or a defending city.  This is mainly used so that
// we can control what objects are passed over DO.  It also means that
// we can use the same method for doing combat among units as well as
// defending cities.
//----------------------------------------------------------------------

typedef struct _CombatProfile
{
    EMMapLocation location;
    Player player;
    int attacking_efficiency;
    int defending_efficiency;
    int damage_per_hit;
    int current_hit_points;
    int adjacent_players;
    BOOL is_a_city;
} CombatProfile;

//----------------------------------------------------------------------
// The WarStatistics gathers together all the information needed for
// the War Report, so that only one method call is required to get the
// information.
//----------------------------------------------------------------------

typedef struct _WarStatistics
{
    int world_city_count;
    int city_count;
    int percent_explored;
    int destroyed_units[UNIT_TYPE_COUNT]; // enemy units destroyed
    int lost_units[UNIT_TYPE_COUNT];
    int under_construction[UNIT_TYPE_COUNT];
    int in_combat[UNIT_TYPE_COUNT];
    int soonest_complete[UNIT_TYPE_COUNT];
} WarStatistics;

//----------------------------------------------------------------------
// The movement type is generally used by the game manager and various
// orders to determine if a unit can move on a type of terrain.
//----------------------------------------------------------------------

typedef short MovementType;

#define LAND_UNIT  0x01
#define WATER_UNIT 0x02
#define AIR_UNIT   (LAND_UNIT|WATER_UNIT)

//----------------------------------------------------------------------
// These bitfields are used to encode the detection capabilities of
// each unit type, as well as the corresponding detectable
// characteristics of each unit.
//
// The game manager uses these to decide what units each player can see
// while exploring.
//----------------------------------------------------------------------

#define VIS_AIR  0x01
#define VIS_SEA  0x02
#define VIS_LAND 0x04
#define VIS_SUB  0x08

//----------------------------------------------------------------------
// Functions for manipulating MapTokens and their components
//----------------------------------------------------------------------

MapToken EMCreateMapToken (Terrain terrain, Player player, Icon icon, BOOL explored);
void EMMapTokenComponents (MapToken mapToken, Terrain *terrain, Player *player, Icon *icon, BOOL *explored);

Terrain EMTerrainComponent (MapToken mapToken);
Player EMPlayerComponent (MapToken mapToken);
Icon EMIconComponent (MapToken mapToken);
BOOL EMExploredComponent (MapToken mapToken);

MapToken EMChangeTerrainComponent (MapToken mapToken, Terrain terrain);
MapToken EMChangePlayerComponent (MapToken mapToken, Player player);
MapToken EMChangeIconComponent (MapToken mapToken, Icon icon);
MapToken EMChangeExploredComponent (MapToken mapToken, BOOL explored);

UnitType EMConvertIconToUnitType (Icon icon);
Icon EMConvertUnitTypeToBaseIcon (UnitType unitType);

//----------------------------------------------------------------------
// Other utility functions for Empire types
//----------------------------------------------------------------------

NSString *EMUnitTypeName (UnitType aType);
NSString *EMShortDirectionName (Direction dir);
NSString *EMLongDirectionName (Direction dir);

BOOL EMUnitTypeCanSee (UnitType viewer, UnitType viewed);
int EMDistanceBetweenLocations (EMMapLocation location1, EMMapLocation location2);

//----------------------------------------------------------------------
// Functions to assist in debugging
//----------------------------------------------------------------------

NSString *EMFormatComponents (MapToken mapToken);
NSString *EMFormatNineComponents (MapToken *tokens);
NSString *EMFormat25Components (MapToken *tokens);

NSString *EMFormatVisibilityFlags (int flags);
NSString *EMFormatNineVisibilityFlags (int *flags);
NSString *EMFormat25VisibilityFlags (int *flags);

//======================================================================
// Notifications
//======================================================================

extern NSString *EMPlayerHasWonGameNotification;
extern NSString *EMGameStateChangedNotification;
extern NSString *EMTurnChangedNotification;
extern NSString *EMPlayerStateChangedNotification;
